############################################
# Takes as input clean sources and substitutes all real declaration with RPE variables.
# Produces a vault with the info on these new sources produced.
############################################
import AutoRPE.UtilsRPE.SourceManager as SourceManager
from AutoRPE.UtilsRPE.ImplementRPE import Implement_RPE
from AutoRPE.UtilsRPE.Error import PathType

from os import mkdir
from os.path import isdir
from shutil import rmtree


def get_command_line_arguments():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input-sources", required=True, type=PathType(exists=True, type='dir'),
                        help="Path to a folder that contains the preprocessed cleaned code source")
    parser.add_argument("-o", "--output-sources", default="SourcesWithRPE",
                        help="Path to the folder that will contain the AutoRPE implementation ")
    parser.add_argument("--output-vault", default="vault.pkl",
                        help="Filename of the output database")

    args = parser.parse_args()

    return args


if __name__ == "__main__":

    args = get_command_line_arguments()

    path_to_input_sources = args.input_sources
    path_to_output_sources = args.output_sources
    output_vault = args.output_vault

    # Paths to source folders
    if isdir(path_to_output_sources):
        rmtree(path_to_output_sources)
    mkdir(path_to_output_sources)

    print("Starting the implementation of the reduced precision emulator in:\n   %s\n\n" % path_to_input_sources)

    # List of files that should not be processed, as red from "list_of_files_to_keep_unmodified.txt"
    preprocess_blacklist = SourceManager.load_list_of_files_to_keep_unmodified()

    # Format text + replace real declarations with type(rpe_var)
    vault = Implement_RPE(path_to_input_sources, path_to_output_sources, preprocess_blacklist)

    vault.save(output_vault)
