import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.Vault as Vault
import AutoRPE.UtilsRPE.TrackDependency as TrackDependency
from os.path import abspath


def get_command_line_arguments():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input-source", default=None, required=True,
                        help="Path to input fortran files")
    parser.add_argument("-o", "--output-vault", default="vault.pkl",
                        help="Name of the output vault file")
    parser.add_argument("-w", "--working-precision", default=23, type=int,
                        help="The precision used for real numbers in the parsed code")

    args = parser.parse_args()

    return args


if __name__ == "__main__":

    args = get_command_line_arguments()

    path_to_input_sources = args.input_source
    output_vault = args.output_vault
    wp = args.working_precision

    # Set the proper working precision
    VariablePrecision.wp = VariablePrecision.precision2type[wp]
    # Parse input sources and stores info in vault
    print("Creating a database from sources found in:\n %s" % path_to_input_sources)
    vault = Vault.Vault(path_to_input_sources)

    # Store the dependencies btw variables, and create a graph for visualization
    TrackDependency.propagate_dependencies(vault, "./graph.gexf")

    # Save the vault with the variable precisions updated.
    vault.save(output_vault)
    print("Vault successfully dumped to ", abspath(output_vault))
