import AutoRPE.UtilsRPE.CurrentBlock as CurrentBlock
from AutoRPE.UtilsRPE.ObtainSourceFileInfo import ReadSourceFile
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.Splitter as Splitter
import AutoRPE.UtilsRPE.Getter as Getter
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.Inserter as Inserter
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
from AutoRPE.UtilsRPE.Vault import Vault
import re


def Implement_RPE(path_to_input_sources, path_to_output_sources, blacklist=[]):
    # Changes all the real declaration for RPE declaration
    implementer = ImplementRPE(path_to_input_sources, path_to_output_sources, blacklist)

    # Fix the problem the changes implied
    implementer.fix_sources()

    # "use rpe_emulator" statement is inserted
    Inserter.insert_use_rpe_statement(implementer.vault)

    # Insert precision specification into sources
    Inserter.add_sbits_to_rpe_variables(implementer.vault)

    # Add the read_precisions module into a file (should be accessible for all the other modules)
    Inserter.insert_read_precisions_module_to_file_and_save(path_to_output_sources, implementer.vault)

    # Create the read_precisions.f90 module
    Inserter.create_read_precisions_module(path_to_output_sources, implementer.vault)

    return implementer.vault


class ImplementRPE:
    def __init__(self, path_to_input_sources, path_to_output_sources, blacklist=[]):
        self.file_list = self.load_sources(path_to_input_sources)
        self.output_path = path_to_output_sources
        self.blacklist = blacklist
        # Replaces real declaration with RPE declaration
        self.replace_real_with_RPE_declaration()
        # Parse the sources and obtain all info
        self.vault = Vault(path_to_output_sources)

    def load_sources(self, path_to_input_sources):
        from os import listdir
        from os.path import join
        ord_list = listdir(path_to_input_sources)
        ord_list.sort()
        file_list = [ReadSourceFile(join(path_to_input_sources, filename)) for filename in
                     ord_list if
                     filename.count(".f90") and filename[0][0] != "."]
        return file_list

    def replace_real_with_RPE_declaration(self):
        n_of_file = len(self.file_list)
        for index, f in enumerate(self.file_list):
            if f.filename not in self.blacklist:
                print("\rFormatting file and replacing reals %i/%i  %20s" % (index + 1, n_of_file, f.filename), end="")

                # Replace real declarations with type(rpe_var)
                self._replace_real_declaration_with_RPE(f)

                # Workaround for parameters
                self._replace_parameter_declaration_with_RPE(f)
            f.save_modified_file(self.output_path)

    def _replace_real_declaration_with_RPE(self, source_file):
        # Look at each line for REAL variable declarations and replace it with type(rpe_var)
        for _index, line in enumerate(source_file.lines):
            _match = re.search(RegexPattern.general_real_declaration, line, flags=re.I)
            if _match:
                _match_real_with_prec = re.search(RegexPattern.real_with_precision_declaration, line, flags=re.I)
                if _match_real_with_prec:
                    real_type = _match_real_with_prec.groups()[0]
                    line = line.replace(real_type, "type(rpe_var)")
                    source_file.lines[_index] = line
                else:
                    _match = re.search(RegexPattern.pure_real_declaration, line, flags=re.I)
                    if _match:
                        real_type = _match.groups()[0]
                    else:
                        raise Exception
                    line = line.replace(real_type, "type(rpe_var)")
                    source_file.lines[_index] = line
        source_file.rebuild_text()

    def _replace_parameter_declaration_with_RPE(self, source_file):
        # If there's an assignation whe have to find a workaround
        for _index, line in enumerate(source_file.lines):
            if re.search('^[^\!]*type\(rpe_var\).*::.*=', line, flags=re.I):
                arguments = line.split("::")[1]

                type_of_exception = self._find_kind_of_parameter_exception(arguments)

                # The assignation is made through a real number
                if type_of_exception == 'is_a_number':
                    piece_to_replace = arguments.split("=")[1].strip()
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace
                    source_file.lines[_index] = line.replace(piece_to_replace, replacement)
                # The assignation is made with a RPE variable
                elif type_of_exception == "regular":
                    piece_to_replace = arguments.split("=")[1].strip()
                    # take the real part of the RPE var
                    piece_to_replace_real = " %s%%val " % piece_to_replace
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace_real
                    source_file.lines[_index] = Inserter.replace_variable_exact_match(piece_to_replace, replacement,
                                                                                      line)
                elif type_of_exception == "multiple_variables":
                    for argument in arguments.split(","):
                        piece_to_replace = argument.split("=")[1].strip()
                        replacement = " rpe_var(  52, %s ) " % piece_to_replace
                        line = line.replace(piece_to_replace, replacement)
                    source_file.lines[_index] = line
                elif type_of_exception == "array_initialization":
                    # Keep just the right part of  assignation avoid error when dimensions are specified
                    piece_to_replace = arguments.split("=")[1].strip()
                    replacement = Inserter.add_rpe_to_array_arguments(piece_to_replace)

                    source_file.lines[_index] = line.replace(piece_to_replace, replacement)
                elif type_of_exception == "array_reshape":
                    array = arguments.split("[")[1]
                    array = array.split("]")[0]

                    # List the unique arguments inside the reshape function
                    arguments_in_reshape = list(set([a.strip() for a in array.split(",")]))

                    # Add the rpe_var cast into each one of them
                    for arg in arguments_in_reshape:
                        piece_to_replace = arg.strip()
                        replacement = " rpe_var(  52, %s ) " % piece_to_replace
                        line = line.replace(piece_to_replace, replacement)
                    source_file.lines[_index] = line
                elif type_of_exception == "has_operation":
                    piece_to_replace = arguments.split("=")[1].strip()
                    # Get the elements in operation
                    split_elements = Splitter.split_elements_in_operations(piece_to_replace)
                    replacement = " rpe_var(  52, %s ) " % piece_to_replace
                    for element in split_elements:
                        operand = element.strip()

                        # If is a real
                        if not NatureDeterminer.is_a_number(operand):
                            operands_real_part = " %s%%val " % operand
                            replacement = Inserter.replace_variable_exact_match(operand, operands_real_part,
                                                                                replacement)
                    line = line.replace(piece_to_replace, replacement)
                    source_file.lines[_index] = line
                elif type_of_exception == 'has_intrinsics':

                    # Get the name of the intrinsic to split the line
                    intrinsic_name = NatureDeterminer.has_intrinsics(arguments, True)
                    piece_to_replace = arguments.split(intrinsic_name)[1].strip()

                    # Match the name of the RPE variable inside the parenthesis
                    p = re.search('\((.*?)\)', piece_to_replace)
                    piece_to_replace = p.group(1)

                    # Get the real part of the RPE variable
                    replacement = " %s%%val " % piece_to_replace

                    # Substitute the mach back in the line
                    line = line.replace(piece_to_replace, replacement)
                    piece_to_replace = line.split("=")[1].strip()

                    # Add the cast to RPE var
                    replacement = " rpe_var( 52, %s ) " % piece_to_replace

                    # Save changes
                    source_file.lines[_index] = line.replace(piece_to_replace, replacement)
                    # self.lines[_index] = line
                elif type_of_exception == "already_fixed":
                    continue
        source_file.rebuild_text()

    def fix_sources(self):
        n_of_modules = len(self.vault.modules)
        for index, module in enumerate(self.vault.modules):
            print("\rFixing modules %i/%i  %20s" % (index + 1, n_of_modules, module.module_name), end="")
            # Process intrinsic calls
            self.fix_intrinsic_functions(module)

            # Process subroutine and function calls
            self.fix_subprogram_calls(module)

            self._fix_type_conversion(module)
            self._fix_data_statement(module)

            # Process read and write calls
            self._fix_read_and_write_calls(module)
            # Fix pointer assignations
            PointerDealer.fix_pointer_assignations(module, self.vault)

    def _fix_data_statement(self, module):
        changed_something = False
        current_block = CurrentBlock.CurrentBlock(module.main)
        # If there's an assignation whe have to find a workaround
        for _index, line in enumerate(module.lines):
            match = re.search(r'^( *)data +(.*) +/(.*)/', line, flags=re.I)
            if match:
                spaces, var_name, values = match.groups(0)
                if var_name.count(","):
                    var_type = Getter.get_type_of_contents(var_name.split(",")[0], procedure=current_block.procedure,
                                                           vault=self.vault)
                else:
                    var_type = Getter.get_type_of_contents(var_name, procedure=current_block.procedure,
                                                           vault=self.vault)
                if var_type != "rpe_var":
                    continue
                values = ",".join(["rpe_var(52, %s)" % x for x in values.split(",")])
                module.lines[_index] = "%sdata %s /%s/" % (spaces, var_name, values)
                changed_something = True
            else:
                current_block.in_which_block(line, vault=self.vault)
        if changed_something:
            module.rebuild_text()

    def _fix_type_conversion(self, module):
        current_block = CurrentBlock.CurrentBlock(module.main)
        for line_index, line in enumerate(module.lines):
            called_derived_type = self._call_to_derived_type_constructor(line, current_block.module_name)
            if not called_derived_type:
                # If not, look if the code is entering or exiting a routine block
                current_block.in_which_block(line, self.vault)
            else:
                # Fix constructor call
                constructor_call = BasicFunctions.close_brackets(line[line.index(called_derived_type.name):])
                called_arguments = CallManager.find_call_arguments(line[line.index(called_derived_type.name):])
                if not called_arguments:
                    continue

                fixed_constructor_call = constructor_call
                for argument_index, argument in enumerate(called_arguments):
                    argument_type = Getter.get_type_of_contents(argument, current_block.procedure, self.vault)
                    dummy_argument = called_derived_type.get_variable_by_position(argument_index)
                    if not dummy_argument:
                        raise AssertionError("Variable not found in position: %i" % argument_index)
                    if argument_type == dummy_argument.type:
                        continue
                    elif argument_type in VariablePrecision.real_types and dummy_argument.type == "rpe_var":
                        fixed_constructor_call = Inserter.add_rpe_to_argument(argument, fixed_constructor_call)
                    else:
                        raise AssertionError("Types do not coincide and the case is not considered %s %s" % (
                            argument_type, dummy_argument.type))
                module.lines[line_index] = line.replace(constructor_call, fixed_constructor_call)
        module.rebuild_text()

    def _fix_read_calls(self, module):
        current_block = CurrentBlock.CurrentBlock(module.main)
        pattern = "[^\w](read *\(.*)"
        for line_index, line in enumerate(module.lines):
            matches = re.search(pattern, line, re.IGNORECASE)
            if re.search("write *\(", line, re.I):
                continue
            if matches:
                read_call = BasicFunctions.close_brackets(matches.groups(1)[0])
                # Remove the read call
                read_arguments = line[matches.start(1):].replace(read_call, "")
                # Continue just if arguments follow the call
                if not read_arguments:
                    continue
                replacement = read_arguments.replace(" ", "")
                arguments = Splitter.split_read_arguments(replacement)

                # Check the argument type and replace just if is an rpe
                for argument_index, argument in enumerate(arguments):
                    argument_type = Getter.get_type_of_contents(argument, procedure=current_block.procedure,
                                                                vault=self.vault)
                    if argument_type == "rpe_var":
                        replacement = Inserter.replace_variable_exact_match(argument, argument + "%val", replacement)

                # Rebuilt the line
                # self.lines[line_index] = read_call + " " + read_arguments
                module.lines[line_index] = line.replace(read_arguments, replacement)
            else:
                current_block.in_which_block(line, vault=self.vault)

    def _fix_write_calls(self, module):
        current_block = CurrentBlock.CurrentBlock(module.main)
        pattern = "^[^\!]*\s*write\([^()]*\) (.*)."

        for line_index, line in enumerate(module.lines):
            if re.search("WRITE *\(", line):
                wc = BasicFunctions.close_brackets(line[line.find("WRITE"):])

                line = line.replace(wc, "WRITE()")
            else:
                current_block.in_which_block(line, vault=self.vault)
            matches = re.search(pattern, line.strip(), re.IGNORECASE)
            if matches:
                line_container, arguments = Splitter.split_line_container_and_arguments(module.lines[line_index])
                if not arguments:
                    continue
                mask = BasicFunctions.literal_mask(line)
                for argument_index, argument in enumerate(arguments):
                    if mask[line.index(argument) + 1]:
                        continue
                    if re.match(RegexPattern.array_in_write_statement, argument):
                        # For calls like  WRITE(numout,9402) (ilci(ji,jj),ilcj(ji,jj),ji=il1,il2)
                        firs_el = Splitter.split_array_in_write_arguments(argument)[0]
                        argument_type = Getter.get_type_of_contents(firs_el, procedure=current_block.procedure,
                                                                    vault=self.vault)
                        if argument_type == "rpe_var":
                            raise Error.ExceptionNotManaged("This part of the code does not yet exist")
                        continue
                    argument_type = Getter.get_type_of_contents(argument, procedure=current_block.procedure,
                                                                vault=self.vault)
                    if argument_type == "rpe_var":
                        arguments[argument_index] = "real( %s )" % argument
                replacement = ",".join(arguments)
                module.lines[line_index] = line_container.replace("###TOREPLACE###", replacement)

    def _fix_read_and_write_calls(self, module):
        self._fix_read_calls(module)
        self._fix_write_calls(module)

    def fix_intrinsic_functions(self, module):
        # Fix all the intrinsics function that are not overloaded in the RPE library
        current_block = CurrentBlock.CurrentBlock(module.main)
        for _index, _line in enumerate(module.lines):

            # Cycle through the matches, that can be more than one in a line
            replacements = []
            pieces_to_replace = []
            matches_start_list = []
            mask = BasicFunctions.literal_mask(_line)
            for intrinsic in SourceManager.list_of_intrinsics:

                # Search for the intrinsics
                fun = r"[ ,(+-=*](%s *)\(" % intrinsic

                matches = re.finditer(fun, _line, flags=re.I)
                matches_start = [m.start(1) for m in matches]

                if matches_start:
                    for start in matches_start:
                        if not mask[start]:
                            matches_start_list.append(start)

            if not matches_start_list:
                current_block.in_which_block(_line, self.vault)
                continue
            else:
                matches_start_list.sort(reverse=True)
                for start in matches_start_list:
                    # Save the piece that need change
                    pieces_to_replace.append(_line[start:].strip())

                    # Current index
                    i_match = len(pieces_to_replace) - 1
                    # Remove initial minus sign (-)
                    if pieces_to_replace[i_match][0] in ["-", "("]:
                        pieces_to_replace[i_match] = pieces_to_replace[i_match][1:]
                    # Remove brackets if they surrounds the argument
                    pieces_to_replace[i_match] = BasicFunctions.close_brackets(pieces_to_replace[i_match])
                    replacements.append(pieces_to_replace[i_match])

                    _line = _line[:start] + _line[start:].replace(pieces_to_replace[i_match],
                                                                  "ALREADY_FIXED_" + str(i_match))
                    # Split the calls argument
                    arguments = CallManager.find_call_arguments(pieces_to_replace[i_match])
                    for arg in arguments:
                        if arg.count("ALREADY_FIXED_"):
                            if NatureDeterminer.has_operations(arg):
                                arguments.extend([el for el in Splitter.split_elements_in_operations(arg) if
                                                  not el.count("ALREADY_FIXED_")])
                            continue
                        argument_type = Getter.get_type_of_contents(arg, current_block.procedure, self.vault)
                        if argument_type != 'rpe_var':
                            continue
                        if re.match("^\w+$", arg):
                            replacements[i_match] = Inserter.add_val_to_argument(arg, replacements[i_match])
                        else:
                            replacements[i_match] = Inserter.add_real_to_argument(arg, replacements[i_match])

            for i_match, (piece_to_replace, replacement) in reversed(
                    list(enumerate(zip(pieces_to_replace, replacements)))):
                # Starting from the end, and substituting just first match we avoid re-substituting matches
                _line = _line.replace("ALREADY_FIXED_" + str(i_match), replacement, 1)
            module.lines[_index] = _line

    def _fix_external_call(self, external_call, line):

        temporal_call = BasicFunctions.remove_if_condition(line)

        called_arguments = external_call.arguments
        # In case the procedure is not found, consider it external
        for argument in called_arguments:
            if argument.type == "rpe_var":
                # Matches a single variable
                if re.match("^\w+$", argument.name):
                    temporal_call = Inserter.add_val_to_argument(argument.name, temporal_call)
                else:
                    # All the other cases (operation, hardcoded array, function call, ...) a cast in easier
                    # TODO: if the intent is out we are screwed
                    temporal_call = Inserter.add_real_to_argument(argument.name, temporal_call)
        return temporal_call.strip()

    def _fix_arguments(self, line, subprogram_call):
        for s_call in subprogram_call:
            if not s_call.subprogram.is_external:
                for arg, dummy_a in zip(s_call.arguments, s_call.dummy_arguments):
                    # This argument is coherent
                    if arg.type == dummy_a.type:
                        continue
                    # Inconsistency between used argument and what the call expect
                    if dummy_a.intent == "in":
                        if dummy_a.type == "rpe_var" and arg.type in VariablePrecision.real_types:
                            line = Inserter.add_rpe_to_argument(arg.name, line)
                        elif dummy_a.type in VariablePrecision.real_types and arg.type == "rpe_var":
                            line = Inserter.add_real_to_argument(arg.name, line)
                    # The value of the argument can be modified by the call, if its an rpe pass the val
                    else:
                        if arg.type == "rpe_var" and dummy_a.type in VariablePrecision.real_types:
                            line = Inserter.add_val_to_argument(arg.name, line)
                        else:
                            Error.ExceptionNotManaged("Here we should use temporal variables")
            else:
                line = self._fix_external_call(s_call, line)
        return line

    def fix_subprogram_calls(self, module):
        current_block = CurrentBlock.CurrentBlock(module.main)

        for _index, line in enumerate(module.lines):
            # Returns arguments of subprograms call and corresponding type and dummy
            temporal_call = BasicFunctions.remove_if_condition(line)
            subprogram_call = \
                Getter.get_procedure_inline(temporal_call, self.vault, current_block)
            # If something was found check if can be fixed
            if subprogram_call:
                module.lines[_index] = line.replace(temporal_call, self._fix_arguments(temporal_call, subprogram_call))
            else:
                current_block.in_which_block(line, vault=self.vault)

    def _find_kind_of_parameter_exception(self, arguments):
        if not arguments.count(","):
            argument = arguments.split("=")[1]
            num_type = NatureDeterminer.is_a_number(arguments)
            if num_type:
                return "is_a_number"
            if NatureDeterminer.has_operations(argument):
                type_of_exception = 'has_operation'
            elif NatureDeterminer.has_intrinsics(argument):
                type_of_exception = 'has_intrinsics'
            else:
                type_of_exception = "regular"
        else:
            if arguments.count("=") > 1:
                type_of_exception = "multiple_variables"
            elif arguments.count("(/"):
                type_of_exception = "array_initialization"
            elif arguments.count("RESHAPE(["):
                type_of_exception = "array_reshape"
            elif re.search(r"= *\[.*\]", arguments):
                type_of_exception = "array_reshape"
            elif arguments.count("rpe_var("):
                type_of_exception = "already_fixed"
            else:
                raise TypeError("Unexpected kind of exception")
        return type_of_exception

    def _call_to_derived_type_constructor(self, line, module_name):
        line = BasicFunctions.remove_literals(line)
        matches = re.finditer(RegexPattern.assignation_to_DT, line, flags=re.I)
        if matches:
            for m in matches:
                fun_name = m.groups()[0]
                try:
                    procedure = self.vault.get_procedure_by_name(fun_name, module_name)
                    if isinstance(procedure, BasicStructures.DerivedType):
                        return procedure
                except Error.ProcedureNotFound:
                    pass
        return None
