import AutoRPE.UtilsRPE.SourceManager as SourceManager
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision

import re


def is_function_declaration_statement(line):
    # Remove matches inside write(*,*) statements
    line = BasicFunctions.remove_literals(line)
    # Matches the word functions, excluding the end word
    is_start = re.search(RegexPattern.function_declaration, line, re.MULTILINE | re.IGNORECASE)
    if is_start:
        return True
    return False


def is_end_function_statement(line):
    is_end = re.match(RegexPattern.end_function_declaration, line, re.I)
    if is_end:
        return True
    return False


# Check if a string is a number and returns the type
def is_a_number(string):
    if string.count("="):
        string = string.split("=")[1]

    # Check if is just white spaces
    if not string.strip():
        return False
    else:
        string = string.lower().strip()

    # Remove unary operator if present
    if string[0] == '-' or string[0] == '+':
        string = string[1:]

    name_of_intrinsic = has_intrinsics(string, True)
    if name_of_intrinsic:
        string = string.replace(name_of_intrinsic, "")
        string = string.strip()
        string = string[1:-1]
        string = string.strip()

    # Checks for number in exponential notation
    exp = re.match("^" + RegexPattern.number_exp_DPE_notation + "$", string)
    # Checks for number with precision specified
    real_with_prec = re.match(RegexPattern.real_number_with_precision, string)
    if exp or real_with_prec:
        # In case the number is a hardcoded real with a type specification,
        # check the lookup table to know to which type it corresponds and return it
        try:
            new_type = VariablePrecision.lookup_table[string.split("_")[1]]
        except IndexError:
            return VariablePrecision.wp
        return new_type
    integer_with_prec = re.match(RegexPattern.integer_number_with_precision, string)
    if integer_with_prec:
        return "integer"
    # Check for pure integer and real numbers
    try:
        float(string)
        if string.isdigit():
            return "integer"
        else:
            return VariablePrecision.wp
    except ValueError:
        return False


def is_an_hardcoded_array(string):
    match = re.search(RegexPattern.hardcoded_array, string)
    if match:
        return True
    return False


def has_operations(string):
    import AutoRPE.UtilsRPE.CallManager as CallManager
    operations = ["+", "-", "*", "/"]
    # Exclude key arguments
    # TODO: Deal with "list comprehension" i.e. : ' (/ (zshift + REAL(ji,wp), ji = 1, jpi_crs*jpj_crs) /)'
    # TODO: This approach does not work
    keyword_pattern = "^ *\w+ *=.*"
    if re.match(keyword_pattern, string, re.I):
        return False
    # Check if string is just white spaces
    if not string.strip():
        return False
    else:
        string = string.lower().replace(" ", "")

    if string.count("'") or string.count('"'):
        return False
    if CallManager.is_call_to_function(string):
        return False
    # Remove the unary operators
    if string[0] == '-' or string[0] == '+':
        string = string[1:]

    # Replace char that can cause false positive
    string = string.replace("=-", 'EQUAL_MIN')
    string = string.replace("(/", "BEGIN_VEC(")
    string = string.replace("/)", ")")

    # Substitutes arrays and structures, then deletes useless brackets
    string = remove_arrays_and_structures(string)

    # Check for exponential, regex is needed to be sure is not just one of the variables that ends with 'e'
    exp_DPE = re.search(RegexPattern.number_exp_DPE_notation, string)
    if exp_DPE:
        string = string.replace(exp_DPE.group(), 'EXPONENTIAL_DPE')

    for op in operations:
        if op in string:
            return True
    return False


def has_intrinsics(string, return_intrinsic_name=False, vault=None):

    for key in list_of_key:
        if string.lower().count(key):
            key_match = r",\s*[ ,(+-]%s *= *\w" % key
            string = re.sub(key_match, "", string, re.I)
    # We just need to check the assignation, not the variable name
    if string.count("="):
        string = string.split("=")[1].strip()

    # Split parenthesis opening
    if not string.count("("):
        return False
    if vault is not None:
        list_of_intrinsics = vault.list_of_intrinsics
    else:
        list_of_intrinsics = SourceManager.list_of_intrinsics

    intrinsic_name = ""
    for intrinsic in list_of_intrinsics:
        try:
            intrinsic_name = re.search(r'\s*\b%s\b\s*\(' % intrinsic, string, re.I).group()
            # Remove last parenthesis and leading spaces
            intrinsic_name = intrinsic_name[:-1].strip()
            break
        except AttributeError:
            continue
    if vault:
        try:
            # In NEMO a variable is declared, that shadows an intrinsic (I am really disappointed)
            vault.variables_dictionary[intrinsic_name]
            return False
        except KeyError:
            pass
    if intrinsic_name:
        if return_intrinsic_name:
            return intrinsic_name
        return True
    return False


def strip_from_intrinsic(string, vault):
    import AutoRPE.UtilsRPE.CallManager as CallManager
    intrinsic = has_intrinsics(string, True, vault)
    while intrinsic:
        string = re.sub("\\b"+intrinsic+"\\b", "", string, re.I)
        if BasicFunctions.is_all_inside_parenthesis(string):
            string = BasicFunctions.remove_outer_brackets(string)
        intrinsic = has_intrinsics(string, True)
    called_arg = CallManager.find_call_arguments("(" + string + ")")
    # Check if is a cast with precision
    if len(called_arg) == 2:
        try:
            VariablePrecision.lookup_table[called_arg[1]]
            return called_arg[:1]
        except KeyError:
            pass
    return called_arg


# PRIVATE

# Substitutes arrays and structures, then deletes useless brackets
# Used in has_operations
def remove_arrays_and_structures(string, ret_rep_list=False):
    replacement_index = 0
    arrays = True
    replacement_list = []
    # Checks for all array pattern, and substitute with string
    while arrays:
        arrays = re.findall(RegexPattern.array, string)
        for array in arrays:
            string = string.replace(array, "REPLACEMENT" + str(replacement_index), 1)
            replacement_list.append(array)
            replacement_index += 1

        # Substitute structures with string
        structures = re.findall(r"\w+%\w+", string)
        for struct in structures:
            string = string.replace(struct, "REPLACEMENT" + str(replacement_index))
            replacement_list.append(struct)
            replacement_index += 1
            arrays = True
    if ret_rep_list:
        return string, replacement_list
    return string


list_of_key = ["dim", "mask", "kind", "back"]
