from AutoRPE.UtilsRPE.ReadSourceFile import ReadSourceFile
import AutoRPE.UtilsRPE.BasicStructures as BasicStructures
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
import AutoRPE.UtilsRPE.CallManager as CallManager
import AutoRPE.UtilsRPE.NatureDeterminer as NatureDeterminer
import AutoRPE.UtilsRPE.Finder as Finder
import AutoRPE.UtilsRPE.AttributeParser as AttributeParser
import AutoRPE.UtilsRPE.CurrentBlock as CurrentBlock
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
import AutoRPE.UtilsRPE.Error as Error
from os.path import join
import re
import pickle


def load_source_files_info(path):
    from os import listdir
    ord_list = listdir(path)
    ord_list.sort()
    file_list = [ObtainSourcefileInfo(ReadSourceFile(join(path, filename)).path) for filename in
                 ord_list if
                 filename.count(".f90") and filename[0][0] != "."]
    return file_list


def parse_sources(path_to_model_sources, vault, save_vault=False):

    info_file_obj = load_source_files_info(path_to_model_sources)

    # Going over all the source-files loading the files,
    # parsing the routines and storing both things in dictionaries.
    n_of_files = len(info_file_obj)
    print("\n")
    for index, f in enumerate(info_file_obj):
        print("\rParsing subprograms and variables %i/%i " % (index + 1, n_of_files), end="")
        f.find_subroutines()
        f.find_functions()
        f.find_module_variables()
        # Store modules, module subprogram and variables
        vault.add_module(f.module)
        for procedure in f.objects:
            try:
                vault.procedures_dictionary[procedure.name] += [procedure]
            except KeyError:
                vault.procedures_dictionary[procedure.name] = [procedure]
            vault.procedures.append(procedure)
            vault[f.module_name].procedures.append(procedure)
            if procedure.name == "main":
                vault[f.module_name].main = procedure
                vault[f.module_name].variables.extend(procedure.variables)

            for var in procedure.variables:
                vault.variables.append(var)

            if isinstance(procedure, BasicStructures.SubRoutine):
                vault.subroutines.append(procedure)

            if isinstance(procedure, BasicStructures.Function):
                vault.functions.append(procedure)

    print("\nParsing Interfaces\n")
    # Going over all the source-files parsing the interfaces and storing them in dictionary.
    for f in info_file_obj:
        f.find_interfaces(vault)
        for interface in f.objects:
            if isinstance(interface, BasicStructures.Interface):
                vault.procedures.append(interface)
                try:
                    vault.procedures_dictionary[interface.name] += [interface]
                except KeyError:
                    vault.procedures_dictionary[interface.name] = [interface]
                vault[f.module_name].procedures.append(interface)
                vault.interfaces.append(interface)
        # Find used module and store
        f.find_used_modules(vault)

    # After having all the used parsed, propagate the inheritance, and add public var to available vars
    print("Expanding use statements\n")
    vault.expand_use_statement()

    print("Parsing derived types\n")
    # Going over all the source-files parsing the derived types and storing them in dictionary.
    for f in info_file_obj:
        f.find_types()
        for derived_type in f.objects:
            if isinstance(derived_type, BasicStructures.DerivedType):
                vault[f.module_name].procedures.append(derived_type)
                vault.derived_types.append(derived_type)
                try:
                    vault.procedures_dictionary[derived_type.name] += [derived_type]
                except KeyError:
                    vault.procedures_dictionary[derived_type.name] = [derived_type]

                vault.procedures.append(derived_type)
                for var in derived_type.variables:
                    vault.variables.append(var)

    # Fill a dictionary to speed up search of variables
    vault.fill_variables_dictionary()

    for f in info_file_obj:
        f.find_namelist_variable(vault)
    # Now that vault.derived_types is full add extension to derived types if any
    vault.expand_derived_types()
    # Store the external used data types
    vault.store_external_datastructures()
    # Stores the external call found in the code
    for f in info_file_obj:
        f.find_external_subroutines(vault)
    # Stores the external call to function found in the code
    for f in info_file_obj:
        f.find_external_functions(vault)
    # Passing to pointed objects the actual function/subroutine they points to
    PointerDealer.process_pointers(vault)

    print("number of functions " + str(len(vault.functions)) + "\n" +
          "number or subroutines " + str(len(vault.subroutines)) + "\n" +
          "number or interfaces " + str(len(vault.interfaces)) + "\n" +
          "number or main " + str(len(vault.modules)) + "\n" +
          "number or procedures " + str(len(vault.procedures)) + "\n" +
          "number of variables " + str(len(vault.variables)) + "\n" +
          "number or derived_types " + str(len(vault.derived_types)) + "\n")

    if save_vault:
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        pickle.dump(vault, open("vault.pkl", "wb"))
    return vault


class ObtainSourcefileInfo(ReadSourceFile):
    # Definition: This class is used to load a sourcefile and parse the subroutines contained there
    # Input: string with path to sourcefile
    def __init__(self, _filename):
        super(ObtainSourcefileInfo, self).__init__(_filename)
        # Definition: Initializing the object reading the sourcefile
        # Input: string with path to sourcefile
        # I think that it is safe to collect the information using all the code in lower case.
        # self.text = self.text.lower()
        self.objects = []
        self.module = BasicStructures.Module(_filename, self.text, self.lines)

    # Definition: Given a line number, returns the subroutine name and the dummy arguments (in case they exist)
    # Input: integer with line number
    # Output: tuple with routine name and list of dummy arguments
    def get_dummy_arguments(self, _index):
        _line = self.lines[_index]
        # Workaround to solve the problem that appears when there's a function declaration with (kind=...)
        if _line.lower().count("function "):
            if _line.find("(") < _line.lower().find("function "):
                _line = _line[_line.lower().find("function "):]

        # If the subroutine declaration does not have any parenthesis, this routine does not have any argument
        if not _line.count("("):
            _name = _line.split()[1].strip()
            _dummy_arguments = []
        # If otherwise it has dummy arguments, find them
        elif _line.lower().count("result") or _line.lower().count("bind("):
            _name = _line[:_line.find("(")].split()[-1]
            _dummy_arguments = []
            _dummy_arguments += CallManager.strip_call_arguments(BasicFunctions.close_brackets(_line))
            # I don't know if we should record the result variable with the other variables
            # result_start = _line.lower().find("result")
            # _dummy_arguments += strip_call_arguments(_line[result_start:])
        else:
            _name = _line[:_line.find("(")].split()[-1]
            _dummy_arguments = CallManager.strip_call_arguments(_line)
        _dummy_arguments = [_x for _x in _dummy_arguments if _x.strip()]
        return _name, _dummy_arguments

    # Definition: Given the line numbers where a routine starts and ends,
    # look for variable declarations.
    # Input:
    #       -_subprogram_start: integer with line number where the routine starts
    #       -_subprogram_end: integer with line number where the routine ends
    # Output: list of strings containing declarations
    def find_subprogram_declaration(self, _subprogram_start, _subprogram_end, obj_to_find="::"):
        # Check if there is some statement inside
        # TODO find a better way of parsing this: actually we can take advantage of the fact that
        # Cases in which a routine contains other functions declarations.
        stop_index = [index
                      for index, line in enumerate(self.lines[_subprogram_start:_subprogram_end + 1])
                      if line.count("contains") or line.count("interface")]
        # A declaration found stop the search there
        if stop_index:
            _subprogram_end = _subprogram_start - 1 + stop_index[0]

        # Lines inside the routine
        _subprogram_lines = self.lines[_subprogram_start:_subprogram_end + 1]

        # Lines that are variable declarations
        _all_declaration = [_l.strip() for _l in _subprogram_lines if
                            re.search(obj_to_find, _l, re.I) and not
                            _l.lower().count("allocate")]

        return _all_declaration

    # Definition: Given the routine name, the name of the dummy arguments and the lines containing declarations,
    # this method finds the attributes of each dummy arguments and adds all the information
    # into the object variable "routines".
    # Input:
    #       -routine_name: string with the routine name
    #       -dummy_arguments: list of strings with argument names
    #       -declarations: list of strings with variable declarations
    def add_struct_members(self, file, structure):
        try:
            for obj in file.objects:
                if obj.name == "main":
                    main = obj
                    break
            for member in structure.variables:
                main.add_variable(member)

        except UnboundLocalError:
            print("Main not found for structure %s", structure.name)

    def find_procedure_variables(self, header, routine_name, dummy_arguments, var_declarations, obj_type, used_modules):
        if obj_type == BasicStructures.Function:
            # Retrieve the ret_val name
            ret_type, ret_val = Finder.find_function_type_from_header(header, var_declarations)
            if not ret_val:
                ret_val = routine_name
        elif obj_type == BasicStructures.SubRoutine:
            ret_val = None
        else:
            raise ValueError("Header does not correspond to function or routine")

        ob = obj_type(routine_name, self.module)

        if len(used_modules):
            ob.use_only_variables = [m.lower().split("use")[1].strip() for m in used_modules]
        # Reverse approach
        dummy_arguments = [x.lower() for x in dummy_arguments]
        for declaration in var_declarations:
            pattern = "(.*)::(.*)"
            m = re.search(pattern, declaration)
            # In this case is not a variable declaration
            if declaration.count("contiguous"):
                continue
            _attributes, variable = m.groups()
            # Check if the dummy argument is inside the declaration arguments
            # If it is, take the type, dimension, intent and optional
            _var_type, _dimension, _intent, _optional, \
            _allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(_attributes, variable)

            variable_name = AttributeParser.get_variable_name(variable)

            if len(variable_name) != 1:
                raise AssertionError("Problem identifying variable declaration.")
            variable_name = variable_name[0]
            _v = BasicStructures.Variable(ob, variable_name, _var_type, _dimension, self.module)
            _v.allocatable = _allocatable
            _v.is_parameter = is_parameter
            # If the variable name is the same of the function or is return statement
            if ret_val is not None and _v.name.lower() == ret_val.lower():
                _v.is_retval = True
                ob.type = _var_type
            _v.is_pointer = is_pointer
            if variable_name.lower() in dummy_arguments:
                _v.intent = _intent
                _v.is_optional = _optional
                _v.is_dummy_argument = True
            ob.add_variable(_v)

        # Save the position of the argument
        found_vars = [variable.name.lower() for variable in ob.variables]

        for index, dummy_argument in enumerate(dummy_arguments):
            if dummy_argument in found_vars:
                # TODO: save position of already found vars
                ob.get_variable_by_name(dummy_argument).position = index  # variable.position = index
                continue
                # If it's not found its an implicitly declared variable
            if dummy_argument[0] in "ijklmn":
                var_type = "integer"
            else:
                var_type = VariablePrecision.wp
            dimension = 0
            intent = None
            optional = None
            _v = BasicStructures.Variable(ob, dummy_argument, var_type, dimension, self.module)
            _v.intent = intent
            _v.is_optional = optional
            _v.is_dummy_argument = True
            _v.position = index
            _v.is_parameter = False
            ob.add_variable(_v)

        if obj_type == BasicStructures.Function:
            # In case of function  declaration like INTEGER FUNCTION name() without variables declaration
            if not [var for var in ob.variables if var.is_retval]:
                # Add the return value to functions variables
                _v = BasicStructures.Variable(ob, ret_val, ret_type, 0, self.module)
                _v.is_retval = True
                ob.add_variable(_v)
                # Specify the type of the function
                ob.type = ret_type
        # Save object
        self.objects.append(ob)

    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_subroutines(self):

        # Look into each line
        for _index, line in enumerate(self.lines):
            # Look if routine starts with (recoursive) subroutine
            if re.search("^\s*\w*\s*subroutine", line, re.IGNORECASE):
                if line.lower().strip().find("end") == 0:
                    continue
                subroutine_start = _index
                # Look for end subroutine statements
                subroutine_end = BasicFunctions.find_next_end(self.lines, _index, "end subroutine")

                # Find name and dummy arguments
                routine_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                var_declarations = self.find_subprogram_declaration(subroutine_start, subroutine_end, obj_to_find="::")
                used_modules = self.find_subprogram_declaration(subroutine_start, subroutine_end, obj_to_find="^use\s")
                header = self.lines[subroutine_start]
                # Parse everything and store
                self.find_procedure_variables(header, routine_name, dummy_arguments, var_declarations,
                                              obj_type=BasicStructures.SubRoutine, used_modules=used_modules)

    def find_external_subroutines(self, vault):
        # Store all the external called subroutines
        current_block = CurrentBlock.CurrentBlock(self.module.main)
        for line in self.lines:
            # Check if there is a call to subroutine in line
            line = BasicFunctions.remove_if_condition(line)
            sbr_call = CallManager.call_to_subroutine(line)

            if not sbr_call:
                current_block.in_which_block(line, vault)
            else:
                # Returns a list, but if the code is preprocessed the list will always have length 1
                try:
                    CallManager.get_subroutine(line, current_block.procedure, vault)

                except Error.ProcedureNotFound as unknown_subroutines:
                    for sbr_name in unknown_subroutines.args[0]:
                        try:
                            vault.external_procedures[sbr_name]
                        except KeyError:
                            vault.external_procedures[sbr_name] = BasicStructures.ExternalSubprogram(sbr_name)

    def find_external_functions(self, vault):
        # Store all the external called subroutines
        current_block = CurrentBlock.CurrentBlock(self.module.main)
        for index, line in enumerate(self.lines):
            # Check if there is a call to function in line
            func_call = re.search(RegexPattern.call_to_function, line.replace("  ", " "), re.I)

            if not func_call or line.count("::") or NatureDeterminer.is_function_declaration_statement(line):
                current_block.in_which_block(line, vault)
            else:
                # Check if the function called is external
                try:
                    CallManager.get_function(line, current_block.procedure, vault)

                except Error.ProcedureNotFound as unknown_functions:
                    for func_name in unknown_functions.args[0]:
                        try:
                            vault.external_procedures[func_name]
                        except KeyError:
                            vault.external_procedures[func_name] = BasicStructures.ExternalSubprogram(func_name)

    def find_types(self):
        # Looks into the sourcefile for routine declarations, storing the name, type,
        # dimensions and optionality of the arguments.
        import AutoRPE.UtilsRPE.RegexPattern as RegexPatter
        # Look into each line
        in_type_declaration = False
        types = {}
        type_name = None
        variable_counter = None
        for _index, line in enumerate(self.lines):
            # Look if routine starts with subroutine
            if not in_type_declaration:
                # check if a type has been declared
                m = re.search(RegexPatter.simple_type_declaration, line, re.IGNORECASE)
                if m:
                    in_type_declaration = True
                    variable_counter = 0
                    type_name = m.groups(0)[0]
                    types[type_name] = {}
                    ob = BasicStructures.DerivedType(type_name, self.module)
                    self.objects.append(ob)
                # Case in which a type is extending another type
                match_extend = re.search(RegexPatter.extended_type_declaration, line, re.IGNORECASE)
                if match_extend:
                    in_type_declaration = True
                    variable_counter = 0
                    type_name = match_extend.groups()[2]
                    extended_type = match_extend.groups()[1]
                    types[type_name] = {}
                    ob = BasicStructures.DerivedType(type_name, self.module, extended_type)
                    self.objects.append(ob)

            else:
                # Check if we are exiting a type declaration
                m = re.search(RegexPatter.end_type_statement, line, re.IGNORECASE)
                if m:
                    in_type_declaration = False
                else:
                    pattern = "(.*)::(.*)"
                    m = re.search(pattern, line)
                    if m:
                        attributes, variables = m.groups()
                        var_type, dimension, intent, optional, \
                        allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(attributes, variables)

                        variables = AttributeParser.get_variable_name(variables)
                        for var in variables:
                            types[type_name][var] = var_type
                            v = BasicStructures.Variable(ob, var, var_type, dimension, self.module)
                            v.position = variable_counter
                            v.is_dummy_argument = True
                            v.allocatable = allocatable
                            v.is_parameter = is_parameter
                            v.is_pointer = is_pointer
                            v.is_member_of = type_name
                            ob.add_variable(v)
                            variable_counter += 1
        return types

    def find_module_variables(self):
        import AutoRPE.UtilsRPE.RegexPattern as RegexPattern
        main_routine = BasicStructures.ModuleMain(self.module)

        line_to_check = True
        self.objects.append(main_routine)
        pattern = "(.*)::(.*)"
        for line in self.lines:
            if line_to_check:
                # Check if we are entering a type declaration zone
                is_type_declaration = re.search(RegexPattern.general_type_declaration, line, re.IGNORECASE)
                if is_type_declaration or NatureDeterminer.is_function_declaration_statement(line):
                    line_to_check = False
                    continue
            else:
                # Check if we are exiting a type declaration zone
                is_end_type_declaration = re.search(RegexPattern.end_type_statement, line, re.IGNORECASE)
                if is_end_type_declaration or NatureDeterminer.is_end_function_statement(line):
                    line_to_check = True
                # Even if we are inside or exiting a type declaration zone, we have to jump to next line
                continue

            if line.lower().count("contains"):
                break
            if line.lower().count("end program"):
                break
            match = re.search(pattern, line)
            if match:
                attributes, variables = match.groups()
                var_type, dimension, intent, optional, \
                allocatable, is_pointer, is_parameter = AttributeParser.parse_attributes(attributes, variables)
                variables = AttributeParser.get_variable_name(variables)
                if var_type:
                    for variable in variables:
                        v = BasicStructures.Variable(main_routine, variable, var_type, dimension, self.module)
                        v.allocatable = allocatable
                        v.is_pointer = is_pointer
                        v.is_parameter = is_parameter
                        main_routine.add_variable(v)

    def find_namelist_variable(self, vault):
        if self.text.lower().count('namelist'):
            current_block = CurrentBlock.CurrentBlock(self.module.main)
            for line in self.lines:
                if not line.lower().strip().find("namelist") == 0:
                    current_block.in_which_block(line, vault=vault)
                else:
                    # Retrieve all variables in the namelist and set the module of the namelist
                    namelist_variables = [var.strip() for var in line.split("/")[2].split(",")]
                    for nv in namelist_variables:
                        var = vault.get_variable_by_name(nv, current_block.procedure)
                        var.appears_in_namelist = self.module_name

    # Definition: Looks into the sourcefile for routine declarations, storing the name, type,
    # dimensions and optionality of the arguments.
    def find_functions(self):
        # Look into each line
        for _index, line in enumerate(self.lines):
            if NatureDeterminer.is_function_declaration_statement(line):
                function_start = _index
                # Look for end subroutine statements
                function_end = BasicFunctions.find_next_end(self.lines, _index, "end function")

                # Find name and dummy arguments
                function_name, dummy_arguments = self.get_dummy_arguments(_index)
                # Find declarations
                variables_declarations = self.find_subprogram_declaration(function_start, function_end)
                used_modules = self.find_subprogram_declaration(function_start, function_end, obj_to_find="use")
                # Parse everything and store
                header = self.lines[function_start]
                self.find_procedure_variables(header, function_name, dummy_arguments, variables_declarations,
                                              obj_type=BasicStructures.Function, used_modules=used_modules)


    # Definition: Looks into the sourcefile for interface declarations, storing the name.
    def find_interfaces(self, vault):
        for _index, _line in enumerate(self.lines):
            if _line.strip().lower().find("interface") == 0:
                interface_start = _index
                interface_end = BasicFunctions.find_next_end(self.lines, _index, "end interface")
                try:
                    interface_name = _line.split()[1].strip()
                except IndexError:
                    # interface to an external procedure or a procedure used in a procedure with the DEFERRED attribute
                    continue
                interfaced_procedure = []
                for interface_line in self.lines[interface_start:interface_end + 1]:
                    if interface_line.strip().lower().find("module procedure") == 0:
                        line_procedure = interface_line.replace("module procedure", "")
                        line_procedure = line_procedure.replace("MODULE PROCEDURE", "")
                        line_procedure = [_r.strip() for _r in line_procedure.split(",")]
                        interfaced_procedure += line_procedure
                    elif interface_line.strip().lower().strip().find("subroutine") == 0:
                        subroutine_name = interface_line[:interface_line.find("(")].split()[1]
                        interfaced_procedure += [subroutine_name]

                ob = BasicStructures.Interface(interface_name, self.module)
                for ip in interfaced_procedure:
                    for procedure in vault.procedures_dictionary[ip]:
                        ob.add_procedure(procedure)
                    if not ob.subprogram:
                        raise AssertionError("No routine found for interface %s" % interface_name)
                self.objects.append(ob)

    def find_used_modules(self, vault):
        current_block = CurrentBlock.CurrentBlock(self.module.main)
        for line in self.lines[1:]:
            current_block.in_which_block(line, vault=vault)
            line = line.strip().lower()
            # No need to parse the entire file
            if not line.find("use ") == 0:
                break
            module_name = line.split()[1]

            use_only_procedure = []
            if line.count(","):
                use_only_procedure = line.split(":")[1].lower().replace("only", "")
                use_only_procedure = use_only_procedure.lower().replace(":", "").strip()
                use_only_procedure = use_only_procedure.split(",")
            if module_name.count(","):
                # Remove the only information
                module_name = module_name.split(",")[0].strip()

            # Sometimes use statement are repeated
            if module_name not in vault[self.module_name].uses:
                # Add to the module the used modules
                module = vault[module_name]
                # If the module is None it means it is an external import
                if module is not None:
                    vault[self.module_name].uses.append(vault[module_name])

            # Look for
            pattern = "(\w+) *=> *(\w+)"
            matches = re.findall(pattern, line)
            if matches:
                for match in matches:
                    pointer, target = match
                    target_var = vault.module.uses[-1].main().get_variable_by_name(target)
                    new_var = BasicStructures.Variable(current_block.procedure, pointer, target_var.type,
                                                       target_var.dimension)
                    current_block.procedure.add_variable(new_var)
