from AutoRPE.UtilsRPE.ObtainSourceFileInfo import ReadSourceFile
import AutoRPE.UtilsRPE.BasicFunctions as BasicFunctions
from AutoRPE.UtilsRPE.Splitter import split_semicolon
import AutoRPE.UtilsRPE.CallManager as CallManager
import re


def load_source_files(path):
    from os import listdir
    from os.path import join
    ord_list = listdir(path)
    ord_list.sort()
    file_list = [Preprocessor(ReadSourceFile(join(path, filename)).path) for filename in
                 ord_list if
                 filename.count(".f90") and filename[0][0] != "."]
    return file_list


# Remove comments, empty lines and old style declaration
def preprocess_sources(path_to_input_sources, path_to_output_sources):
    all_sources = load_source_files(path_to_input_sources)
    n_of_file = len(all_sources)
    if len(all_sources):
        for index, f in enumerate(all_sources):
            print("\rFormatting file %i/%i  %20s" % (index + 1, n_of_file, f.filename), end="")
            f.pre_process()
            f.save_modified_file(path_to_output_sources)
    else:
        # Sources is an empty folder
        raise ValueError("No files found in %s folder", path_to_input_sources)


class Preprocessor(ReadSourceFile):
    def __init__(self, filename):
        super(Preprocessor, self).__init__(filename)

    def pre_process(self, replace_wrk_alloc=False):
        # Remove all comments from all the lines
        self._remove_all_comments()
        # Deletes all the empty lines
        self.lines = BasicFunctions.remove_empty_lines(self.lines)
        # Merges commands split on multiples lines
        self.lines = BasicFunctions.merge_lines(self.lines)
        # Split multiple command lines
        self._split_lines()
        # Split lines with more than one declaration
        self._split_declarations()

        # Just needed for old Nemo version
        if replace_wrk_alloc:
            self.replace_workalloc_in_file()

        # Remove parameter assignation F77 style
        self._remove_F77_parameter_declaration()
        # Put text back together
        self.rebuild_text()

    def _split_lines(self):
        # Split lines which contains more than one command
        # TODO: We should be sure that we are not splitting by characters in literals
        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            if line.count(";"):
                parts = split_semicolon(line)
                self.lines[_index] = parts[0]
                for i in range(1, len(parts)):
                    self.lines.insert(_index + 1, parts[i])
                    length += 1
                    _index += 1
            _index += 1
        self.rebuild_text()

    # Split declarations of variables putting a single variable in each line
    def _split_declarations(self):
        _index = 0
        length = len(self.lines)
        while _index < length:
            line = self.lines[_index]
            line = BasicFunctions.remove_literals(line)
            if line.count("::"):
                attributes, arguments = line.split("::")
                arguments = CallManager.find_call_arguments("c(%s)" % arguments)
                if len(arguments) > 1:
                    new_line = "%s :: %s" % (attributes, arguments[0])
                    self.lines[_index] = new_line
                    for argument in arguments[1:]:
                        new_line = "%s :: %s" % (attributes, argument)
                        self.lines.insert(_index + 1, new_line)
                        length += 1
                        _index += 1

            _index += 1
        self.rebuild_text()

    # Remove comments from a file
    def _remove_all_comments(self):
        for _index, _line in enumerate(self.lines):
            self.lines[_index] = BasicFunctions.remove_comments(_line)

    # If parameter declaration is made in Fortran 77 style
    def _remove_F77_parameter_declaration(self):

        pattern = '^\s*PARAMETER\s*\(\s*[\w+=\s*]'

        for line_index, line in enumerate(self.lines):

            # Search for lines with the 'PARAMETER' statement
            match = re.match(pattern, line, re.I)
            if match:
                param_line = line.replace(")", "")

                # Store parameter names
                parameters_name = [parameter.split("=")[0].strip() for parameter in
                                   param_line.split("(")[1].split(",")]

                # Store parameter values
                parameters_val = [parameter.split("=")[1].strip() for parameter in
                                  param_line.split("(")[1].split(",")]

                for parameter_name, parameter_value in zip(parameters_name, parameters_val):

                    # Search for each parameter declaration backward in the file
                    for dec_line_index, dec_line in enumerate(reversed((self.lines[:line_index]))):
                        parameter_pattern = '.*::\s*' + parameter_name + '$'
                        param_declaration = re.match(parameter_pattern, dec_line, re.I)
                        if (line_index - dec_line_index - 1) == 0:
                            raise AssertionError("declaration of parameter %s not found in file %s" % (
                                parameter_name, self.filename))
                        # Put back in place the parameter attribute and assign the parameter vale
                        if param_declaration:
                            # Add param statement
                            piece_to_replace = dec_line.split('::')[0]
                            replacement = "%s, parameter " % piece_to_replace
                            dec_line = dec_line.replace(piece_to_replace, replacement)

                            # Add param assignation
                            dec_line = dec_line + "=" + parameter_value

                            # Substitute the line
                            self.lines[line_index - dec_line_index - 1] = dec_line

                            break

                # Delete parameter declaration
                self.lines[line_index] = ""
