# Stores pattern reused along the code

# Capures a real number with precision specification: 0.5_dp / 0_sp
real_number_with_precision = r"^[0-9\.]*_[a-zA-Z]+$"
integer_number_with_precision = r"^[0-9\.]*_\d+$"
# Pattern for numbers in exponential form, or numbers with double precision exponent
number_exp_DPE_notation = r'\s*\d+\.?\d*[e|d][+|-]*\d+[\w]*'

#### ARRAY
# Matches simple array, and sliced array
array = r"\w+\s*\(\s*[\w,:\d\s+\-\*=]+\s*\)"
# array(:)%b
array_with_member = array + r"%\w+"
#  array(:)%b(:)
array_with_array_member = array + "%" + array

###### STRUCTURE
# Simple strucure a%b
structure = r"\w+%\w+"
# Simple structure with array member a%b(:)
structure_with_member_array = r"\w+%" + array

# Matches with the (/old, style, vec /) or the new [array, style]
# hardcoded_array = r"(\(\/|\[)(\s*[\w\s,%.()\+\-=*]*\s*)(\/\)|\])"
hardcoded_array = r"^\s*(\(\/|\[)(.*)(\/\)|\])"
# In write statement there can be a statement like
# WRITE(numout,9402) jj, (ilci(ji,jj), ilcj(ji,jj),ji=il1,il2)
# the second argument has the type of the first el of the array
array_in_write_statement = r"^\s*\([\s\w,\(\):]+=[\s\w,\(\):]+\)"

#### DECLARATION

# TYPE

# Captures simple types declaration
simple_type_declaration = "^ *type[^()]*[ \t](\w+) *$"
# Captures just types declaration with extended the attribute
extended_type_declaration = "^ *type\s*,(\s*public\s*,|)\s*extends *\(\s*(\w*) *\) *:: * (\w*)"
# Captures type with possible public and extended attributes
general_type_declaration = '^ *type\s*(,|)\s*(public\s*::|public|)\s*(,|)\s*(extends\((\w*)\)\s*::|) * (\w*)'
# Captures end of groups
end_type_statement = "^ *END TYPE"

# REAL (Can be a variable declaration, or a function ret_val specification)

# All the lines that declare a real
general_real_declaration = '^\s*(real).*(::|(function))'
# real(wp/sp/dp/[...]) :: var
real_with_precision_declaration = '[^\!]*(real *\([\w= ]*\)).*(::|(function))'
# real :: var
pure_real_declaration = '^\s*\w?(real).*(::|(function))'

# POINTER

# Captures when a pointer is assigned on a line and gives back
# - group 0 => eventual ptr attributes
# - group 2 => ptr
# - group 3 => target
pointers_assignation = "(.*::|)\s*(.*)=>\s*(.*)"

# SUBPROGRAM

function_declaration = r'function(?!(?<=\bend function)\s*)[^_]'
end_function_declaration = r'^\s*end\s*function'
subprogram_declaration = r'(subroutine|function)\s+\w+\s*\(.*\)'

# CALL
call_to_function = r"(?<!subroutine )(?<!call )[+-]*(\b[a-z_]+[\d]*[a-z_]*)\s*\("
# call_to_function = r"[-+]*\s*(\w+)\s*\("
call_to_subroutine = r"^\s*call\s+(\w+)"
assignation_to_DT = r"(\w+) *\("


### Allocation

variable_allocation = r"allocate *\(.*[^\w]%s[^\w]"