from __future__ import print_function
from os.path import dirname, abspath, isfile
import re


# Get all files contained in a source path (recursive)
def expand_sources(source_path):
    from os import walk
    from os.path import join
    import re
    pattern = ".*\.[fFh90]"
    files = [join(dp, f) for dp, dn, fn in walk(source_path) for f in fn]
    files = [f for f in files if re.search(pattern, f)]
    # Sorting the list to solve the random-like behaviour of walk
    files = list(sorted(files))
    return files


# Gets the filepath of the corresponding module
def match_module_and_filename(module, filenames):
    import re
    pattern = ".*/%s\.[fFh90]" % module
    matches = [fn for fn in filenames if re.match(pattern, fn)]
    return matches


def load_variable_list(filename):
    # Load list of variables
    from numpy import loadtxt
    variables = loadtxt(filename, dtype=int).tolist()
    if not isinstance(variables, list):
        variables = [variables]
    return variables


def load_file(filename):
    with open(filename, "r", encoding="latin-1") as f:
        # Read, strip and keep only lines with content
        _lines = [_l.strip() for _l in f if _l.strip()]
    return _lines

# Hardcoded list, this should be solved
list_of_intrinsics = ["max", "maxloc", "maxval", "min", "minloc", "minval", "epsilon",
                      "isnan", "count", "abs", "mod", "reshape",
                      "atanh", "sqrt", "log",
                      "sum",
                      "erfc",
                      "dsqrt", "cmplx", "transfer"]


def load_intrinsics():
    package_directory = dirname(abspath(__file__))
    # Hardcoded path to file.
    with open(package_directory + "/../AdditionalFiles/fortran_intrinsics") as f:
        return [re.search(r"(\w+)", l).groups(0)[0].lower() for l in f]


def load_list_of_files_to_keep_unmodified(filename=None):
    if filename is None:
        package_directory = dirname(abspath(__file__))
        filename = package_directory + "/../AdditionalFiles/list_of_files_to_keep_unmodified.txt"
    # Hardcoded path to file.
    if isfile(filename):
        print("Loading list of exceptions from %s\n\n" % filename)
        with open(filename) as f:
            return [l.strip() for l in f if l.strip() and l.strip()[0] != "#"]
    else:
        return None


def store_variable_list(list_of_variables, filename, vault):
    # Save list of variables into a txt file
    # Sort and keep only uniq values
    list_of_variables = list(sorted(set(list_of_variables)))
    line_format = "%5i    # %20s %25s %20s\n"
    with open(filename, "w") as output_file:
        # Add header
        output_file.write("#%5s     %20s %25s %20s\n%s\n" % ("ID", "variable", "routine", "module", 100 * "#"))
        # Add variables
        for var_id in list_of_variables:
            v = vault.get_variable_by_id(var_id)
            output_file.write(line_format % (v.id, v.name, v.procedure.name, v.procedure.module.name))


def store_lines_into_file(_lines, filename):
    """
       Saves a list of lines into a file with a given filename.
       """
    to_write = "\n".join(_lines).strip()
    with open(filename, "w") as f:
        f.write(to_write)
