import AutoRPE.UtilsRPE.CurrentBlock as CurrentBlock
import AutoRPE.UtilsRPE.Getter as Getter
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
import networkx as nx


def create_nodes(variables, graph):
    for v in variables:
        try:
            graph[var_identifier(v)]
        except KeyError:
            dim = v.dimension if v.dimension is not None else 0
            graph.add_node(var_identifier(v),
                           dimension=dim,
                           routine=str(v.procedure.name),
                           module=str(v.procedure.module.name),
                           intent=str(v.intent),
                           type=str(v.type))


def create_edges(argument, dummy, graph):
    v1, v2 = argument, dummy
    dim = v1.dimension if v1.dimension is not None else 0
    graph.add_edge(var_identifier(v1), var_identifier(v2),
                   routine=str(v2.procedure.name),
                   kind=str(v2.intent),
                   weight=1 + dim / 4.
                   )


def track_dependencies(vault, graph):

    total = len(vault.modules)

    for index, module in enumerate(vault.modules):
        print("\rTracing dependencies %i/%i  %20s" % (index + 1, total, module.name), end="")

        current_block = CurrentBlock.CurrentBlock(module.main)

        for line in module.lines:
            # Returns all REAL var on the line and corresponding dummy argument
            var, dummy_var = Getter.get_real_in_line(line, vault, current_block)
            if [var, dummy_var] != [None, None]:
                for v, dv in zip(var, dummy_var):
                    # Create nodes
                    create_nodes([v, dv], graph)
                    # Add edges
                    create_edges(v, dv, graph)
            else:
                current_block.in_which_block(line, vault)


def propagate_dependencies(vault, graph_name):
    real_variables = [v for v in vault.variables if v.type in VariablePrecision.real_types]

    # Assigning id and precision
    for index, var in enumerate(real_variables):
        var.id = index
        var.precision = VariablePrecision.type2precision[var.type]

    # Creates a graph to be filled
    graph = nx.DiGraph()

    # Track dependencies and stores the info in vault
    track_dependencies(vault, graph)

    nx.readwrite.gexf.write_gexf(graph,  graph_name)
    variables_in_graph = [v for v in vault.variables if var_identifier(v) in graph.nodes]
    var_dict = {}

    for v in variables_in_graph:
        var_dict[var_identifier(v)] = v
    undirected_graph = graph.to_undirected()

    for v in variables_in_graph:
        cc = nx.node_connected_component(undirected_graph, var_identifier(v))
        try:
            v.same_as = [var_dict[n] for n in cc]
        except KeyError:
            raise KeyError


def var_identifier(var):
    return "%s#%s#%s#%s" % (var.name, var.procedure.name, var.procedure.module.name, str(var.id))
